
# This file was automatically generated by Dist::Zilla::Plugin::ModuleBuild v6.024.
use strict;
use warnings;

use Module::Build 0.28;


my %module_build_args = (
  "build_requires" => {
    "Module::Build" => "0.28"
  },
  "configure_requires" => {
    "Module::Build" => "0.28"
  },
  "dist_abstract" => "Perl module to download/fetch and save information from the Web",
  "dist_author" => [
    "Ian Kluft <https://github.com/ikluft>"
  ],
  "dist_name" => "WebFetch",
  "dist_version" => "0.15.0",
  "license" => "gpl",
  "module_name" => "WebFetch",
  "recursive_test_files" => 1,
  "requires" => {
    "Carp" => 0,
    "Class::Singleton" => 0,
    "DB_File" => 0,
    "Data::Dumper" => 0,
    "Date::Calc" => 0,
    "DateTime" => 0,
    "DateTime::Format::ISO8601" => 0,
    "DateTime::Locale" => 0,
    "Exception::Class" => 0,
    "Getopt::Long" => 0,
    "HTTP::Request" => 0,
    "LWP::UserAgent" => 0,
    "Readonly" => 0,
    "Scalar::Util" => 0,
    "Try::Tiny" => 0,
    "base" => 0,
    "strict" => 0,
    "utf8" => 0,
    "warnings" => 0
  },
  "test_requires" => {
    "File::Basename" => 0,
    "File::Compare" => 0,
    "File::Slurp" => 0,
    "File::Temp" => 0,
    "IO::Capture::Stderr" => 0,
    "String::Interpolate::Named" => 0,
    "Test::Exception" => 0,
    "Test::More" => 0,
    "YAML::XS" => 0,
    "autodie" => 0,
    "open" => 0
  }
);


my %fallback_build_requires = (
  "File::Basename" => 0,
  "File::Compare" => 0,
  "File::Slurp" => 0,
  "File::Temp" => 0,
  "IO::Capture::Stderr" => 0,
  "Module::Build" => "0.28",
  "String::Interpolate::Named" => 0,
  "Test::Exception" => 0,
  "Test::More" => 0,
  "YAML::XS" => 0,
  "autodie" => 0,
  "open" => 0
);


unless ( eval { Module::Build->VERSION(0.4004) } ) {
  delete $module_build_args{test_requires};
  $module_build_args{build_requires} = \%fallback_build_requires;
}

my $build = Module::Build->new(%module_build_args);


$build->create_build_script;
